/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2009                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#include<vcg/simplex/vertex/base.h>
#include<vcg/simplex/vertex/component.h>

#include <vcg/complex/used_types.h>

#include<vcg/simplex/face/base.h>
#include<vcg/simplex/face/component.h>
#include<vcg/complex/complex.h>
#include<wrap/io_trimesh/import_ply.h>
#include<wrap/io_trimesh/export_ply.h>
#include<vcg/space/index/grid_static_ptr.h>
#include<vcg/complex/algorithms/closest.h>
#include<vcg/complex/algorithms/update/normal.h>
#include<vcg/complex/algorithms/update/bounding.h>
using namespace vcg;
using namespace std;

class MyEdge;
class MyFace;
class MyVertex;
struct MyUsedTypes : public UsedTypes<	Use<MyVertex>   ::AsVertexType,
                                        Use<MyEdge>     ::AsEdgeType,
                                        Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public Vertex<MyUsedTypes, vertex::Coord3f, vertex::Normal3f, vertex::BitFlags  >{};
class MyFace    : public Face< MyUsedTypes, face::VertexRef,  face::Normal3f, face::Mark, face::FFAdj, face::BitFlags > {};
class MyEdge    : public Edge<MyUsedTypes>{};
class MyMesh    : public tri::TriMesh< vector<MyVertex>, vector<MyFace> , vector<MyEdge>  > {};


bool ImproveTestDiagonal(MyFace &f, int z)
{
  Point3f od0 = f.V0(z)->cP();
  Point3f od1 = f.V1(z)->cP();
  Point3f nd0 = f.V2(z)->cP();
  Point3f nd1 = f.FFp(z)->V2(f.FFi(z))->cP();

  if(Distance(od0,od1) > Distance(nd0,nd1)) return true;

  return false;
}

float SmallestAngle(Point3f &p0,Point3f &p1,Point3f &p2)
{
  float a0 = Angle(p1-p0,p2-p0);
  float a1 = Angle(p2-p1,p0-p1);
  float a2 = Angle(p0-p2,p1-p2);


  return math::Min(a0,a1,a2);
}

// Return true if the smallest angle become larger
bool ImproveTestSmallestAngle(MyFace &f, int z)
{
  Point3f od0 = f.V0(z)->cP();
  Point3f od1 = f.V1(z)->cP();
  Point3f nd0 = f.V2(z)->cP();
  Point3f nd1 = f.FFp(z)->V2(f.FFi(z))->cP();

  float o0 = SmallestAngle(od0,od1,nd0);
  float o1 = SmallestAngle(od1,od0,nd1);

  float n0 = SmallestAngle(nd0,nd1,od0);
  float n1 = SmallestAngle(nd1,nd0,od1);

  if(min(o0,o1) < min(n0,n1)) return true;
  return false;
}

void ImproveMesh(MyMesh &m)
{
  int changedTri =0;
  tri::UpdateTopology<MyMesh>::FaceFace(m);
  do
  {
    changedTri=0;
    tri::UpdateFlags<MyMesh>::FaceClearV(m);
    for(MyMesh::FaceIterator fi = m.face.begin() ;fi != m.face.end();++fi)
    {
      if(!(fi->IsV()))
      {
        for(int i=0;i<3;i++)
        {
          if(Index(m,*fi)==25 && Index(m, fi->FFp(i))==72)
          {
            printf("25-72\n");
          }
          bool normalTest = face::CheckFlipEdgeNormal(*fi,i,math::ToRad(1.0));
          bool topoTest = face::CheckFlipEdge(*fi,i);
          bool qualityTest = ImproveTestSmallestAngle(*fi,i);
          if(normalTest && topoTest && qualityTest)
          {
            fi->SetV();
            fi->FFp(i)->SetV();
            face::FlipEdge(*fi,i);
            changedTri++;
          }
        }
      }
    }
  }
  while(changedTri>0);
}
bool SnapToVertex(MyFace &f, Point3f &qp, float threshold)
{
  for(int i=0;i<3;++i)
  {
    if(Distance(f.V(i)->cP(),qp)<threshold)
    {
      qp=f.V(i)->cP();
      return true;
    }
  }
  return false;
}

bool SnapToEdge(MyFace &f, Point3f &qp, int &edge, float threshold)
{
  Point3f bar;
  InterpolationParameters(f,qp,bar);
  int minInd;

  if(bar[0]<bar[1])
  {
    if(bar[0]<bar[2]) minInd=0;
    else minInd=2;
  }
  else
  {
    if(bar[1]<bar[2]) minInd=1;
    else minInd=2;
  }

  bar[(minInd+1)%3] += bar[minInd]/2.0f;
  bar[(minInd+2)%3] += bar[minInd]/2.0f;
  bar[minInd]=0;

  Point3f np=f.V(0)->cP()*bar[0] + f.V(1)->cP()*bar[1] + f.V(2)->cP()*bar[2];
  if(Distance(np,qp)<threshold)
  {
    qp=np;
    edge = (minInd+1)%3;
    return true;
  }
  return false;
}

int main( int argc, char **argv )
{
  MyMesh bm, pm, om0,om1;

  if(argc < 3 )
  {
    printf("trimeshbase basemesh.ply pointmesh.ply\n");
    return -1;
  }

  printf("Reading %s\n",argv[1]);

  int ret= tri::io::ImporterPLY<MyMesh>::Open(bm,argv[1]);
  if(ret!=0) {
    printf("Unable to open %s for '%s'\n",argv[1],tri::io::ImporterPLY<MyMesh>::ErrorMsg(ret));
    return -1;
  }
  ret = tri::io::ImporterPLY<MyMesh>::Open(pm,argv[2]);
  if(ret!=0) {
    printf("Unable to open %s for '%s'\n",argv[2],tri::io::ImporterPLY<MyMesh>::ErrorMsg(ret));
    return -1;
  }
  printf("BaseMesh has %i vert and %i faces\n",bm.vn, bm.fn);
  printf("PointMesh has %i vert and %i faces\n",pm.vn, pm.fn);

  int insertedNum=0;
  tri::UpdateFlags<MyMesh>::VertexClearV(pm);
  tri::UpdateBounding<MyMesh>::Box(bm);
  float maxdist = bm.bbox.Diag()/100.0;
  const float snapThr = bm.bbox.Diag()/200.0;
  float founddist;
  Point3f cp;

  MyMesh *curM = &bm;
  MyMesh *outM = &om0;
  int step=0;
  while(insertedNum<pm.vn)
  {
    assert(outM != curM);
    GridStaticPtr<MyFace> myGrd;
    printf("\nIteration %i \n",step);
    printf("Mesh to refine has %i faces\n",curM->fn);
    tri::UpdateNormals<MyMesh>::PerFaceNormalized(*curM);
    tri::UpdateBounding<MyMesh>::Box(*curM);
    tri::UpdateTopology<MyMesh>::FaceFace(*curM);
    myGrd.Set(curM->face.begin(),curM->face.end(),100);

    tri::FaceTmark<MyMesh> markerFunctor(curM);
    face::PointDistanceBaseFunctor<float> PDistFunct;

    tri::Allocator<MyMesh>::AddVertices(*outM,curM->vn);
    for(size_t i=0;i<curM->vn;++i)
      outM->vert[i].P() = curM->vert[i].cP();

    tri::UpdateFlags<MyMesh>::FaceClearV(*curM);
    // Main loop over all the points to be inserted
    for(size_t i=0;i<pm.vn;++i) if(!pm.vert[i].IsV())
    {
      Point3f qp = pm.vert[i].cP(); // search the face nearest to the the i-th vertex of the point mesh;
      MyMesh::FacePointer fp=myGrd.GetClosest(PDistFunct,markerFunctor,qp,maxdist,founddist,cp);
      assert(fp); // we have built our point set with point onto the surface so it should always find a correct face
      if(!(fp->IsV()))  // we copy and split un-visited faces onto a new mesh;
        {
         int edgeInd;
         if(SnapToVertex(*fp,qp,snapThr))
         {
           pm.vert[i].SetV();
           pm.vert[i].P()=qp;
//           printf("Snapped vertex\n");
           insertedNum++;
         }
         else
         if(SnapToEdge(*fp,qp,edgeInd,snapThr))
         {
           if(!fp->FFp(edgeInd)->IsV())
           {
//           printf("snap vertex %i into face %i-%i\n",i,
//                  tri::Index(*curM,fp),tri::Index(*curM,fp->FFp(edgeInd)));
           pm.vert[i].P()=qp;
           fp->SetV();
           fp->FFp(edgeInd)->SetV();
           pm.vert[i].SetV();
           insertedNum++;
           MyFace *newfp = &*tri::Allocator<MyMesh>::AddFaces(*outM,4);
           tri::Allocator<MyMesh>::AddVertices(*outM,1);

           int od0 = tri::Index(*curM,fp->V0(edgeInd));
           int od1 = tri::Index(*curM,fp->V1(edgeInd));
           int nd1 = tri::Index(*curM,fp->V2(edgeInd));
           int nd0 = tri::Index(*curM,fp->FFp(edgeInd)->V2(fp->FFi(edgeInd)));
//           printf("   Vertexes %i %i %i %i\n",od0,od1,nd0,nd1);
           outM->vert.back().P()=qp;
           int nv = outM->vn-1;

           newfp[0].V(0)=&outM->vert[nd0];
           newfp[0].V(1)=&outM->vert[od1];
           newfp[0].V(2)=&outM->vert[nv];

           newfp[1].V(0)=&outM->vert[od1];
           newfp[1].V(1)=&outM->vert[nd1];
           newfp[1].V(2)=&outM->vert[nv];

           newfp[2].V(0)=&outM->vert[nd1];
           newfp[2].V(1)=&outM->vert[od0];
           newfp[2].V(2)=&outM->vert[nv];

           newfp[3].V(0)=&outM->vert[od0];
           newfp[3].V(1)=&outM->vert[nd0];
           newfp[3].V(2)=&outM->vert[nv];

//           for(int i=0;i<4;++i)
//             printf("tri (%i) %i %i %i\n",tri::Index(*outM,newfp[i]),
//                    tri::Index(*outM,newfp[i].V(0)),tri::Index(*outM,newfp[i].V(1)),tri::Index(*outM,newfp[i].V(2)));
         }
         }
         else
         {
           printf("inserting vertex %i into face %i\n",i,tri::Index(*curM,fp));
          fp->SetV();
          pm.vert[i].SetV();
          insertedNum++;
          int v0 = tri::Index(*curM,fp->V(0));
          int v1 = tri::Index(*curM,fp->V(1));
          int v2 = tri::Index(*curM,fp->V(2));

          MyFace *fp = &*tri::Allocator<MyMesh>::AddFaces(*outM,3);
          tri::Allocator<MyMesh>::AddVertices(*outM,1);

          outM->vert.back().P()=qp;
          int nv = outM->vn-1;

          fp[0].V(0)=&outM->vert[v0];
          fp[0].V(1)=&outM->vert[v1];
          fp[0].V(2)=&outM->vert[nv];

          fp[1].V(0)=&outM->vert[v1];
          fp[1].V(1)=&outM->vert[v2];
          fp[1].V(2)=&outM->vert[nv];

          fp[2].V(0)=&outM->vert[v2];
          fp[2].V(1)=&outM->vert[v0];
          fp[2].V(2)=&outM->vert[nv];
         }
        }
    }
    printf("Inserted %i vertices\n",insertedNum);
    int insertedFaceNum=0;
    // Second loop: add all the non splitted faces
    for(MyMesh::FaceIterator fi = curM->face.begin() ;fi != curM->face.end();++fi)
      if(!fi->IsV())
      {
        insertedFaceNum++;
        MyFace *fp = &*tri::Allocator<MyMesh>::AddFaces(*outM,1);
        for(int jj=0;jj<3;++jj)
          fp->V(jj) = &outM->vert[ tri::Index(*curM,fi->V(jj))];
      }
    char buf[128];
    sprintf(buf,"out%02i.ply",step);
    tri::io::ExporterPLY<MyMesh>::Save(*outM,buf,false);
    ImproveMesh(*outM);
    sprintf(buf,"out%02i_f.ply",step);
    tri::io::ExporterPLY<MyMesh>::Save(*outM,buf,false);

    curM = outM;
    outM = (curM==&om0)?&om1:&om0 ;
    outM->Clear();
    step++;

  }

  tri::io::ExporterPLY<MyMesh>::Save(*outM,"out.ply");
  return 0;
}
