/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2009                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#include<vcg/simplex/vertex/base.h>
#include<vcg/simplex/vertex/component.h>

#include <vcg/complex/used_types.h>

#include<vcg/simplex/face/base.h>
#include<vcg/simplex/face/component.h>
#include<vcg/complex/complex.h>
#include<wrap/io_trimesh/import_ply.h>
#include<wrap/io_trimesh/export_ply.h>
#include<vcg/space/index/grid_static_ptr.h>
#include<vcg/complex/algorithms/closest.h>
#include<vcg/complex/algorithms/update/normal.h>
#include<vcg/complex/algorithms/update/bounding.h>
using namespace vcg;
using namespace std;

class MyEdge;
class MyFace;
class MyVertex;
struct MyUsedTypes : public UsedTypes<	Use<MyVertex>   ::AsVertexType,
                                        Use<MyEdge>     ::AsEdgeType,
                                        Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public Vertex<MyUsedTypes, vertex::Coord3f, vertex::Normal3f, vertex::BitFlags  >{};
class MyFace    : public Face< MyUsedTypes, face::VertexRef,  face::Normal3f, face::Mark, face::FFAdj, face::BitFlags > {};
class MyEdge    : public Edge<MyUsedTypes>{};
class MyMesh    : public tri::TriMesh< vector<MyVertex>, vector<MyFace> , vector<MyEdge>  > {};


float FlipImprovement(MyFace &f,int i)
{
  MyVertex *OldDiag0 = f.V0(i);
  MyVertex *OldDiag1 = f.V1(i);

  MyVertex *NewDiag0 = f.V2(i);
  MyVertex *NewDiag1 = f.FFp(i)->V2(f.FFi(i));

  assert((NewDiag1 != NewDiag0) && (NewDiag1 != OldDiag0) && (NewDiag1 != OldDiag1));

  return Distance(OldDiag0->cP(),OldDiag1->cP())/Distance(NewDiag0->cP(),NewDiag1->cP());
}

void FlipOptimizer(MyMesh &m,
                   float thr)
{

  int flipNum=0;
  MyMesh::FaceIterator fi;
  float AngleThrRad = math::ToRad(5.0f);

  tri::UpdateTopology<MyMesh>::FaceFace(m);
  do
  {
    flipNum=0;
    tri::UpdateFlags<MyMesh>::FaceClearV(m);
    tri::UpdateNormals<MyMesh>::PerFaceNormalized(m);

    for(fi=m.face.begin();fi!=m.face.end();++fi)
    {
      for(int i=0;i<3;++i)
      {
        MyMesh::FacePointer ffp=fi->FFp(i);
        if(!fi->IsV() && !ffp->IsV())
        {
          bool goodFlip = face::CheckFlipEdge(*fi,i);
          bool goodNormalFlip = face::CheckFlipEdgeNormal(*fi,i,AngleThrRad);
          if(goodFlip && goodNormalFlip )
          {
            float incr = FlipImprovement(*fi,i);
            if(incr > thr)
            {
              fi->SetV();
              ffp->SetV();
              face::FlipEdge(*fi,i);
              printf("flipping Face %i %i (%5.2f",Index(m,*fi),Index(m,ffp),incr);
              flipNum++;
            }
          }
        }
      }
    } // end foreach face
  }
  while(flipNum >0);



}


int main( int argc, char **argv )
{
  MyMesh bm, pm, om0,om1;

  if(argc < 3 )
  {
    printf("trimeshbase basemesh.ply pointmesh.ply\n");
    return -1;
  }

  printf("Reading %s\n",argv[1]);

  int ret= tri::io::ImporterPLY<MyMesh>::Open(bm,argv[1]);
  if(ret!=0) {
    printf("Unable to open %s for '%s'\n",argv[1],tri::io::ImporterPLY<MyMesh>::ErrorMsg(ret));
    return -1;
  }
  ret = tri::io::ImporterPLY<MyMesh>::Open(pm,argv[2]);
  if(ret!=0) {
    printf("Unable to open %s for '%s'\n",argv[2],tri::io::ImporterPLY<MyMesh>::ErrorMsg(ret));
    return -1;
  }
  printf("BaseMesh has %i vert and %i faces\n",bm.vn, bm.fn);
  printf("PointMesh has %i vert and %i faces\n",pm.vn, pm.fn);

  int insertedNum=0;
  tri::UpdateFlags<MyMesh>::VertexClearV(pm);
  float maxdist = bm.bbox.Diag()/100.0;
  float founddist;
  Point3f cp;

  MyMesh *curM = &bm;
  MyMesh *outM = &om0;
  int step=0;
  while(insertedNum<pm.vn)
  {
    assert(outM != curM);
    GridStaticPtr<MyFace> myGrd;
    printf("\nIteration %i \n",step);
    printf("Mesh to refine has %i faces\n",curM->fn);
    tri::UpdateNormals<MyMesh>::PerFaceNormalized(*curM);
    tri::UpdateBounding<MyMesh>::Box(*curM);
    myGrd.Set(curM->face.begin(),curM->face.end(),100);

    tri::FaceTmark<MyMesh> markerFunctor(curM);
    face::PointDistanceBaseFunctor<float> PDistFunct;

    tri::Allocator<MyMesh>::AddVertices(*outM,curM->vn);
    for(size_t i=0;i<curM->vn;++i)
      outM->vert[i].P() = curM->vert[i].cP();

    tri::UpdateFlags<MyMesh>::FaceClearV(*curM);
    for(size_t i=0;i<pm.vn;++i) if(!pm.vert[i].IsV())
    {
      Point3f qp = pm.vert[i].cP(); // search the face nearest to the the i-th vertex of the point mesh;
      MyMesh::FacePointer fp=myGrd.GetClosest(PDistFunct,markerFunctor,qp,maxdist,founddist,cp);
      assert(fp); // we have built our point set with point onto the surface so it should always find a correct face
      if(!(fp->IsV()))  // we copy and split un-visited faces onto a new mesh;
        {
        //printf("inserting vertex %i into face %i\n",i,tri::Index(*curM,fp));

        pm.vert[i].SetV();
        insertedNum++;
          fp->SetV();
          int v0 = tri::Index(*curM,fp->V(0));
          int v1 = tri::Index(*curM,fp->V(1));
          int v2 = tri::Index(*curM,fp->V(2));

          MyFace *fp = &*tri::Allocator<MyMesh>::AddFaces(*outM,3);
          tri::Allocator<MyMesh>::AddVertices(*outM,1);

          outM->vert.back().P()=qp;
          int nv = outM->vn-1;

          fp[0].V(0)=&outM->vert[v0];
          fp[0].V(1)=&outM->vert[v1];
          fp[0].V(2)=&outM->vert[nv];

          fp[1].V(0)=&outM->vert[v1];
          fp[1].V(1)=&outM->vert[v2];
          fp[1].V(2)=&outM->vert[nv];

          fp[2].V(0)=&outM->vert[v2];
          fp[2].V(1)=&outM->vert[v0];
          fp[2].V(2)=&outM->vert[nv];
        }
    }
    printf("Inserted %i vertices\n",insertedNum);
    int insertedFaceNum=0;
    for(MyMesh::FaceIterator fi = curM->face.begin() ;fi != curM->face.end();++fi) if(!fi->IsV())
    {
      insertedFaceNum++;
      MyFace *fp = &*tri::Allocator<MyMesh>::AddFaces(*outM,1);
      for(int jj=0;jj<3;++jj)
          fp->V(jj) = &outM->vert[ tri::Index(*curM,fi->V(jj))];
    }
    char buf[128];
    sprintf(buf,"out%02i.ply",step);
    tri::io::ExporterPLY<MyMesh>::Save(*outM,buf,false);
    FlipOptimizer(*outM,1.1f);
    sprintf(buf,"out%02i_f.ply",step);
    tri::io::ExporterPLY<MyMesh>::Save(*outM,buf,false);

    curM = outM;
    outM = (curM==&om0)?&om1:&om0 ;
    outM->Clear();
    step++;
  }

  tri::io::ExporterPLY<MyMesh>::Save(*outM,"out.ply");
  return 0;
}
